import os
import logging
import numpy as np
import unittest
from pathlib import Path

from bgen_reader import example_filepath
from bgen_reader import open_bgen
from bgen_reader._multimemmap import MultiMemMap
from pysnptools.util import log_in_place
import shutil
import math
import subprocess
import pysnptools.util as pstutil
from pysnptools.distreader import DistReader
from bed_reader import get_num_threads


def default_iid_function(sample):
    """
    The default function for turning a Bgen sample into a
    two-part :attr:`pysnptools.distreader.DistReader.iid`.
    If the Bgen sample contains a single comma, we split on the comma to
    create the iid.
    Otherwise, the iid will be ('0',sample)

    >>> default_iid_function('fam0,ind0')
    ('fam0', 'ind0')
    >>> default_iid_function('ind0')
    ('0', 'ind0')
    """
    fields = sample.split(",")
    if len(fields) == 2:
        return fields[0], fields[1]
    else:
        return ("0", sample)


def default_sid_function(id, rsid):
    """
    The default function for turning a Bgen (SNP) id and rsid into a
    :attr:`pysnptools.distreader.DistReader.sid`.
    If the Bgen rsid is '' or '0', the sid will be the (SNP) id.
    Otherwise, the sid will be 'ID,RSID'

    >>> default_sid_function('SNP1','rs102343')
    'SNP1,rs102343'
    >>> default_sid_function('SNP1','0')
    'SNP1'
    """
    if rsid == "0" or rsid == "":
        return id
    else:
        return id + "," + rsid


def default_sample_function(famid, indid):
    """
    The default function for turning a two-part :attr:`pysnptools.distreader.DistReader.iid` into a a Bgen sample.
    If the iid's first part (the family id) is '0' or '', the sample will be iid's 2nd part.
    Otherwise, the sample will be 'FAMID,INDID'

    >>> default_sample_function('fam0','ind0')
    'fam0,ind0'
    >>> default_sample_function('0','ind0')
    'ind0'
    """
    if famid == "0" or famid == "":
        return indid
    else:
        return famid + "," + indid


def default_id_rsid_function(sid):
    """
    The default function for turning a :attr:`pysnptools.distreader.DistReader.sid` into a Bgen (SNP) id and rsid.
    If the sid contains a single comma, we split on the comma to create the id and rsid.
    Otherwise, the (SNP) id will be the sid and the rsid will be '0'

    >>> default_id_rsid_function('SNP1,rs102343')
    ('SNP1', 'rs102343')
    >>> default_id_rsid_function('SNP1')
    ('SNP1', '0')
    """
    fields = sid.split(",")
    if len(fields) == 2:
        return fields[0], fields[1]
    else:
        return sid, "0"


class Bgen(DistReader):
    """
    A :class:`.DistReader` for reading \*.bgen files from disk.

    See :class:`.DistReader` for general examples of using DistReaders.

    The BGEN format is described `here <https://www.well.ox.ac.uk/~gav/bgen_format/>`__.

        **Tip:** The 'gen' in BGEN stands for 'genetic'. The 'gen' in :class:`DistGen` stands for generate, because it generates random (genetic) data.*

    **Constructor:**
        :Parameters: * **filename** (string) -- The BGEN file to read.
                     * **iid_function** (optional, function) -- Function to turn a BGEN sample into a :attr:`DistReader.iid`.
                       (Default: :meth:`bgen.default_iid_function`.)
                     * **sid_function** (optional, function or string) -- Function to turn a BGEN (SNP) id and rsid into a :attr:`DistReader.sid`.
                       (Default: :meth:`bgen.default_sid_function`.) Can also be the string 'id' or 'rsid', which is faster than using a function.
                     * **sample** (optional, string) -- A GEN sample file. If given, overrides information in \*.bgen file.
                     * **num_threads** (optinal, int) -- The number of threads with which to read data. Defaults to all available processors.
                            Can also be set with these environment variables (listed in priority order):
                            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
                     * **fresh_properties** (optional, bool) -- When true (default), memory will be allocated for the iid, sid, and
                       pos properties. This is safe. When false, the properties will use the Bgen's on-disk 'memmap'. That saves
                       memory, but is only safe if the Bgen object is still around when the properties are used.

        :Example:

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/example.bgen")
        >>> data_on_disk = Bgen(bgen_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 199)

    **Methods beyond** :class:`.DistReader`

    """

    def __init__(
        self,
        filename,
        iid_function=default_iid_function,
        sid_function=default_sid_function,
        sample=None,
        num_threads=None,
        fresh_properties=True,
    ):
        super(Bgen, self).__init__()
        self._ran_once = False

        self.filename = str(filename)
        self._iid_function = iid_function
        self._sid_function = sid_function
        self._sample = sample
        self._num_threads = num_threads
        self._fresh_properties = fresh_properties

    @property
    def row(self):
        self._run_once()
        if self._fresh_properties and not self._row.flags["OWNDATA"]:
            self._row = self._row.copy()
        return self._row

    @property
    def col(self):
        self._run_once()
        if self._fresh_properties and not self._col.flags["OWNDATA"]:
            self._col = self._col.copy()
        return self._col

    @property
    def col_property(self):
        """*same as* :attr:`pos`"""
        self._run_once()
        if self._fresh_properties and not self._col_property.flags["OWNDATA"]:
            self._col_property = self._col_property.copy()
        return self._col_property

    def _apply_iid_function(self, samples):
        assert len(samples) > 0, "Expect at least one sample"
        if self._iid_function is not default_iid_function:
            if not self._fresh_properties:
                logging.info(
                    "'fresh_properties' is False, but non-default iid_function is given, so allocating memory"
                )
            return np.array(
                [self._iid_function(sample) for sample in samples], dtype="str"
            )
        else:
            return self._open_bgen._metadata2_memmaps[self._default_iid_key]

    def _apply_sid_function(self, id_list, rsid_list):
        if self._sid_function == "id":
            return id_list
        elif self._sid_function == "rsid":
            return rsid_list
        elif self._sid_function is default_sid_function:
            return self._open_bgen._metadata2_memmaps[self._default_sid_key]
        else:
            logging.info("Allocating memory for non-default sid_function")
            return np.array(
                [self._sid_function(id, rsid) for id, rsid in zip(id_list, rsid_list)],
                dtype="str",
            )

    _col_property_key = "_pysnptools_col_property"
    _default_iid_key = "_pysnptools_default_iid"
    _default_sid_key = "_pysnptools_default_sid"

    def _run_once(self):
        if self._ran_once:
            return

        assert os.path.exists(self.filename), "Expect file to exist ('{0}')".format(
            self.filename
        )
        verbose = logging.getLogger().level <= logging.INFO

        self._open_bgen = open_bgen(self.filename, self._sample, verbose=verbose)
        assert (
            self._open_bgen.nvariants == 0 or self._open_bgen.nalleles[0] == 2
        ), "expect number of alleles to be 2"
        assert (
            self._open_bgen.nvariants == 0 or not self._open_bgen.phased[0]
        ), "expect data to be unphased"

        if self._col_property_key not in self._open_bgen._metadata2_memmaps:
            logging.info("Extending metadata file with PySnpTools metadata")
            assert (
                self._default_iid_key not in self._open_bgen._metadata2_memmaps
                and self._default_sid_key not in self._open_bgen._metadata2_memmaps
            ), "real assert"
            metadata2_temp = self._open_bgen._metadata2_path.parent / (
                self._open_bgen._metadata2_path.name + ".temp"
            )
            if metadata2_temp.exists():
                metadata2_temp.unlink()
            metadata2_path = self._open_bgen._metadata2_path
            del self._open_bgen
            shutil.copy(metadata2_path, metadata2_temp)
            with MultiMemMap(metadata2_temp, mode="r+") as metadata2_memmaps:
                samples = metadata2_memmaps["samples"]
                row = metadata2_memmaps.append_empty(
                    self._default_iid_key,
                    shape=(len(samples), 2),
                    dtype=str(samples.dtype),
                )
                if len(samples) == 0 or "," not in samples[0]:
                    logging.info(
                        "No comma in first sample, so extending metadata file with 'no-comma' default iids"
                    )
                    row[:, 0] = "0"
                    row[:, 1] = samples
                else:
                    # Later: Do this in chunks of size about len(samples) using old np.stack(np.core.defchararray.split(samples,',',maxsplit=2)) code (see pre 8/1/2020 code)
                    with log_in_place(
                        "splitting samples on commas", logging.INFO
                    ) as updater:
                        for i, val in enumerate(samples):
                            if i % 1000 == 0:
                                updater(f"{i:,} of {len(samples):,}")
                            row[i, :] = default_iid_function(val)
                col_property = metadata2_memmaps.append_empty(
                    self._col_property_key,
                    shape=(len(metadata2_memmaps["ids"]), 3),
                    dtype="float",
                )
                col_property[:, 2] = metadata2_memmaps["positions"]
                # Do something fast if all numbers
                try:
                    col_property[:, 0] = metadata2_memmaps["chromosomes"]
                except Exception:
                    # If that doesn't work, do something slow
                    with log_in_place(
                        "converting chromosomes strings to numbers, one at a time",
                        logging.INFO,
                    ) as updater:
                        for i, val in enumerate(metadata2_memmaps["chromosomes"]):
                            if i % 1000 == 0:
                                updater(f"{i:,} of {len(col_property):,}")
                            try:
                                col_property[i, 0] = int(val)
                            except Exception:
                                col_property[i, 0] = 0

                rsid_list = metadata2_memmaps["rsids"]
                id_list = metadata2_memmaps["ids"]
                assert str(rsid_list.dtype).startswith("<U") and str(
                    id_list.dtype
                ).startswith("<U"), "real assert"
                if _all_equal_in_parts(rsid_list, "0") or _all_equal_in_parts(
                    rsid_list, ""
                ):
                    col = metadata2_memmaps.append_empty(
                        self._default_sid_key,
                        shape=len(metadata2_memmaps["ids"]),
                        dtype=str(id_list.dtype),
                    )
                    col[:] = id_list
                else:
                    max_length = (
                        int(str(rsid_list.dtype)[2:]) + 1 + int(str(id_list.dtype)[2:])
                    )
                    col = metadata2_memmaps.append_empty(
                        self._default_sid_key,
                        shape=len(metadata2_memmaps["ids"]),
                        dtype=f"<U{max_length}",
                    )
                    for _, _, start, end in _parts(len(col)):
                        col[start:end] = np.char.add(
                            np.char.add(id_list[start:end], ","), rsid_list[start:end]
                        )

            metadata2_path.unlink()
            shutil.copy(metadata2_temp, metadata2_path)
            self._open_bgen = open_bgen(self.filename, self._sample, verbose=verbose)
        else:
            assert (
                self._default_iid_key in self._open_bgen._metadata2_memmaps
                and self._default_sid_key in self._open_bgen._metadata2_memmaps
            ), "real assert"

        self._row = self._apply_iid_function(self._open_bgen.samples)
        self._col = self._apply_sid_function(self._open_bgen.ids, self._open_bgen.rsids)
        self._col_property = self._open_bgen._metadata2_memmaps[self._col_property_key]
        self._assert_iid_sid_pos(check_val=False)
        self._ran_once = True

    def _read(
        self,
        iid_index_or_none,
        sid_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        self._run_once()

        if order == "A":
            order = "F"

        num_threads = get_num_threads(
            self._num_threads if num_threads is None else num_threads
        )

        val = self._open_bgen.read(
            (iid_index_or_none, sid_index_or_none),
            dtype=dtype,
            order=order,
            num_threads=num_threads,
        )
        assert val.shape[-1] == 3, "Expect ploidy to be 2"
        return val

    def __repr__(self):
        return "{0}('{1}')".format(self.__class__.__name__, self.filename)

    def __del__(self):
        self.flush()

    def flush(self):
        """Close the \*.bgen file for reading. (If values or properties are accessed again, the file will be reopened.)

        >>> from pysnptools.distreader import Bgen
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/example.bgen")
        >>> data_on_disk = Bgen(bgen_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 199)
        >>> data_on_disk.flush()

        """
        if hasattr(self, "_ran_once") and self._ran_once:
            self._ran_once = False
            if (
                hasattr(self, "_open_bgen") and self._open_bgen is not None
            ):  # we need to test this because Python doesn't guarantee that __init__ was fully run
                del self._open_bgen

    @staticmethod
    def write(
        filename,
        distreader,
        bits=16,
        compression=None,
        sample_function=default_sample_function,
        id_rsid_function=default_id_rsid_function,
        iid_function=default_iid_function,
        sid_function=default_sid_function,
        block_size=None,
        qctool_path=None,
        cleanup_temp_files=True,
    ):
        """Writes a :class:`DistReader` to BGEN format and return a the :class:`.Bgen`. Requires access to the 3rd party QCTool.

        :param filename: the name of the file to create
        :type filename: string
        :param distreader: The data that should be written to disk. It can also be any distreader, for example, :class:`.DistNpz`, :class:`.DistData`, or
           another :class:`.Bgen`.
        :type distreader: :class:`DistReader`
        :param bits: Number of bits, between 1 and 32 used to represent each 0-to-1 probability value. Default is 16.
            An np.float32 needs 23 bits. A np.float64 would need 52 bits, which the BGEN format doesn't offer, so use 32.
        :type bits: int
        :param compression: How to compress the file. Can be None (default), 'zlib', or 'zstd'.
        :type compression: string
        :param sample_function: Function to turn a :attr:`DistReader.iid` into a BGEN sample.
           (Default: :meth:`bgen.default_sample_function`.)
        :type sample_function: function
        :param id_rsid_function: Function to turn a  a :attr:`DistReader.sid` into a BGEN (SNP) id and rsid.
           (Default: :meth:`bgen.default_id_rsid_function`.)
        :type id_rsid_function: function
        :param iid_function: Function to turn a BGEN sample into a :attr:`DistReader.iid`.
           (Default: :meth:`bgen.default_iid_function`.)
        :type iid_function: function
        :param sid_function: Function to turn a BGEN (SNP) id and rsid into a :attr:`DistReader.sid`.
           (Default: :meth:`bgen.default_sid_function`.)
        :type sid_function: function
        :param block_size: The number of SNPs to read in a batch from *distreader*. Defaults to a *block_size* such that *block_size* \* *iid_count* is about 100,000.
        :type block_size: number
        :param qctool_path: Tells the path to the 3rd party `QCTool <https://www.well.ox.ac.uk/~gav/qctool_v2/>`_. Defaults to reading
           path from environment variable QCTOOLPATH. (To use on Windows, install Ubuntu for Windows, install QCTool in Ubuntu,
           and then give the path as "ubuntu run <UBUNTU PATH TO QCTOOL".)
        :type qctool_path: string
        :param cleanup_temp_files: Tells if delete temporary \*.gen and \*.sample files.
        :type cleanup_temp_files: bool
        :rtype: :class:`.Bgen`

        >>> from pysnptools.distreader import DistHdf5, Bgen
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> hdf5_file = example_file("pysnptools/examples/toydata.snpmajor.dist.hdf5")
        >>> distreader = DistHdf5(hdf5_file)[:,:10] # A reader for the first 10 SNPs in Hdf5 format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.bgen")
        >>> Bgen.write("tempdir/toydata10.bgen",distreader)        # Write data in BGEN format
        Bgen('tempdir/toydata10.bgen')
        """
        qctool_path = qctool_path or os.environ.get("QCTOOLPATH")
        assert (
            qctool_path is not None
        ), "Bgen.write() requires a path to an external qctool program either via the qctool_path input or by setting the QCTOOLPATH environment variable."

        # We need the +1 so that all three values will have enough precision to be very near 1
        # The max(3,..) is needed to even 1 bit will have enough precision in the gen file
        genfilename = os.path.splitext(filename)[0] + ".gen"
        decimal_places = max(3, math.ceil(math.log(2**bits, 10)) + 1)
        Bgen.genwrite(
            genfilename,
            distreader,
            decimal_places,
            id_rsid_function,
            sample_function,
            block_size,
        )

        dir, file = os.path.split(filename)
        if dir == "":
            dir = "."
        metadata_mmm = open_bgen._metadata_path_from_filename(
            file, samples_filepath=None
        )
        samplefile = os.path.splitext(file)[0] + ".sample"
        genfile = os.path.splitext(file)[0] + ".gen"
        olddir = os.getcwd()
        os.chdir(dir)

        if os.path.exists(file):
            os.remove(file)
        if os.path.exists(metadata_mmm):
            os.remove(metadata_mmm)
        cmd = "{0} -g {1} -s {2} -og {3}{4}{5}".format(
            qctool_path,
            genfile,
            samplefile,
            file,
            " -bgen-bits {0}".format(bits) if bits is not None else "",
            " -bgen-compression {0}".format(compression)
            if compression is not None
            else "",
        )
        try:
            _ = subprocess.check_output(
                cmd, stderr=subprocess.STDOUT, shell=True, universal_newlines=True
            )
        except subprocess.CalledProcessError as exc:
            print(
                "Status : FAIL", exc.returncode, exc.output
            )  # LATER this doesn't seem to work when called from VS
            if cleanup_temp_files:
                raise Exception("qctool command failed")
            else:
                print("qctool command failed\n{0}".format(cmd))
        if cleanup_temp_files:
            os.remove(genfile)
            os.remove(samplefile)
        os.chdir(olddir)
        new_bgen = Bgen(filename, iid_function=iid_function, sid_function=sid_function)
        return new_bgen

    @staticmethod
    def genwrite(
        filename,
        distreader,
        decimal_places=None,
        id_rsid_function=default_id_rsid_function,
        sample_function=default_sample_function,
        block_size=None,
    ):
        """Writes a :class:`DistReader` to Gen format

        :param filename: the name of the file to create (will also create *filename_without_gen*.sample file.)
        :type filename: string
        :param distreader: The data that should be written to disk. It can also be any distreader, for example, :class:`.DistNpz`, :class:`.DistData`, or
           another :class:`.Bgen`.
        :type distreader: :class:`DistReader`
        :param decimal_places: (Default: None) Number of decimal places with which to write the text numbers. *None* writes to full precision.
        :type bits: int or None
        :param id_rsid_function: Function to turn a  a :attr:`DistReader.sid` into a GEN (SNP) id and rsid.
           (Default: :meth:`bgen.default_id_rsid_function`.)
        :type id_rsid_function: function
        :param sid_function: Function to turn a GEN (SNP) id and rsid into a :attr:`DistReader.sid`.
           (Default: :meth:`bgen.default_sid_function`.)
        :type sid_function: function
        :param block_size: The number of SNPs to read in a batch from *distreader*. Defaults to a *block_size* such that *block_size* \* *iid_count* is about 100,000.
        :type block_size: number
        :rtype: None

        >>> from pysnptools.distreader import DistHdf5, Bgen
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> hdf5_file = example_file("pysnptools/examples/toydata.snpmajor.dist.hdf5")
        >>> distreader = DistHdf5(hdf5_file)[:,:10] # A reader for the first 10 SNPs in Hdf5 format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.bgen")
        >>> Bgen.genwrite("tempdir/toydata10.gen",distreader)        # Write data in GEN format
        """
        # https://www.cog-genomics.org/plink2/formats#gen
        # https://web.archive.org/web/20181010160322/http://www.stats.ox.ac.uk/~marchini/software/gwas/file_format.html

        block_size = block_size or max((100 * 1000) // max(1, distreader.row_count), 1)

        if decimal_places is None:

            def format(num):
                return f"{num}"

            format_function = format
        else:

            def format(num):
                return ("{0:." + str(decimal_places) + "f}").format(num)

            format_function = format

        start = 0
        updater_freq = 10000
        index = -1
        with log_in_place("writing text values ", logging.INFO) as updater:
            with open(filename + ".temp", "w", newline="\n") as genfp:
                while start < distreader.sid_count:
                    distdata = distreader[:, start : start + block_size].read(
                        view_ok=True
                    )
                    for sid_index in range(distdata.sid_count):
                        id, rsid = id_rsid_function(distdata.sid[sid_index])
                        assert id.strip() != "", "id cannot be whitespace"
                        assert rsid.strip() != "", "rsid cannot be whitespace"
                        genfp.write(
                            "{0} {1} {2} {3} A G".format(
                                int(distdata.pos[sid_index, 0]),
                                id,
                                rsid,
                                int(distdata.pos[sid_index, 2]),
                            )
                        )
                        for iid_index in range(distdata.iid_count):
                            index += 1
                            if (
                                updater_freq > 1
                                and index > 0
                                and index % updater_freq == 0
                            ):
                                updater(
                                    "{0:,} of {1:,} ({2:.2%})".format(
                                        index,
                                        distreader.iid_count * distreader.sid_count,
                                        1.0
                                        * index
                                        / (distreader.iid_count * distreader.sid_count),
                                    )
                                )
                            prob_dist = distdata.val[iid_index, sid_index, :]
                            if not np.isnan(prob_dist).any():
                                s = " " + " ".join(
                                    (format_function(num) for num in prob_dist)
                                )
                                genfp.write(s)
                            else:
                                genfp.write(" 0 0 0")
                        genfp.write("\n")
                    start += distdata.sid_count
        sample_filename = os.path.splitext(filename)[0] + ".sample"
        # https://www.well.ox.ac.uk/~gav/qctool_v2/documentation/sample_file_formats.html
        with open(sample_filename, "w", newline="\n") as samplefp:
            samplefp.write("ID\n")
            samplefp.write("0\n")
            for f, i in distreader.iid:
                samplefp.write("{0}\n".format(sample_function(f, i)))

        if os.path.exists(filename):
            os.remove(filename)
        shutil.move(filename + ".temp", filename)

    def copyinputs(self, copier):
        # doesn't need to self.run_once() because only uses original inputs
        copier.input(self.filename)
        if self._sample is not None:
            copier.input(self._sample)
        metadata2 = open_bgen._metadata_path_from_filename(
            self.filename, samples_filepath=self._sample
        )
        if os.path.exists(metadata2):
            copier.input(metadata2)


def _parts(item_count):
    if item_count == 0:
        return
    part_sqrt = int(math.ceil(item_count / math.sqrt(item_count)))
    start = 0
    for part_index in range(part_sqrt):
        end = min(start + part_sqrt, item_count)
        yield part_index, part_sqrt, start, end
        start = end
    assert end == item_count, "real assert"


def _all_equal_in_parts(array, val):
    for _, _, start, end in _parts(len(array)):
        if not np.all(array[start:end] == val):
            return False
    return True


class TestBgen(unittest.TestCase):
    @staticmethod
    def assert_approx_equal(distdata0, distdata1, atol):
        from pysnptools.pstreader import PstData

        assert PstData._allclose(distdata0.row, distdata1.row, equal_nan=True)
        assert PstData._allclose(distdata0.col, distdata1.col, equal_nan=True)
        assert PstData._allclose(
            distdata0.row_property, distdata1.row_property, equal_nan=True
        )
        assert PstData._allclose(
            distdata0.col_property, distdata1.col_property, equal_nan=True
        )
        np.testing.assert_allclose(
            distdata0.val, distdata1.val, atol=atol, equal_nan=True, verbose=True
        )

    def test1(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        logging.info("in TestBgen test1")
        bgen = Bgen("../examples/example.bgen")
        bgen.read()
        bgen2 = bgen[:2, ::3]
        bgen2.read()
        os.chdir(old_dir)

    def test_other(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        logging.info("in TestBgen test_other")
        bgen = Bgen("../examples/example.bgen", sample="../examples/other.sample")
        assert np.all(bgen.iid[0] == ("0", "other_001"))
        os.chdir(old_dir)

    def test_zero(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        bgen = Bgen("../examples/example.bgen")
        assert bgen[:, []].read().val.shape == (500, 0, 3)
        assert bgen[[], :].read().val.shape == (0, 199, 3)
        assert bgen[[], []].read().val.shape == (0, 0, 3)
        os.chdir(old_dir)

    def test2(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        logging.info("in TestBgen test2")
        bgen = Bgen("../examples/bits1.bgen")
        bgen.read()
        bgen2 = bgen[:2, ::3]
        bgen2.read()
        os.chdir(old_dir)

    def test_memmap(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        assert (
            "QCTOOLPATH" in os.environ
        ), "To run test_read_write_round_trip, QCTOOLPATH environment variable must be set. (On Windows, install QcTools in 'Ubuntu on Windows' and set to 'ubuntu run <qctoolLinuxPath>')."

        distgen0data = Bgen("../examples/example.bgen")[:, 10].read()
        assert distgen0data.iid[0, 0] == "0"
        # distgen0data = DistGen(seed=332,iid_count=50,sid_count=5).read()
        file1y = "temp/roundtrip1-{0}-{1}y.bgen".format(0, 1)
        bgen = Bgen.write(file1y, distgen0data)
        assert bgen.iid is not None
        file1x = "temp/roundtrip1-{0}-{1}x.bgen".format(0, 1)
        assert Bgen.write(file1x, distgen0data).iid[0, 0] == "0"
        os.chdir(old_dir)

    def test_read_write_round_trip(self):
        from pysnptools.distreader import DistGen

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        assert (
            "QCTOOLPATH" in os.environ
        ), "To run test_read_write_round_trip, QCTOOLPATH environment variable must be set. (On Windows, install QcTools in 'Ubuntu on Windows' and set to 'ubuntu run <qctoolLinuxPath>')."

        exampledata = Bgen("../examples/example.bgen")[:, 10].read()
        distgen0data = DistGen(seed=332, iid_count=50, sid_count=5).read()

        for i, distdata0 in enumerate([distgen0data, exampledata]):
            for bits in list(range(1, 33)):
                logging.info("input#={0},bits={1}".format(i, bits))
                file1 = "temp/roundtrip1-{0}-{1}.bgen".format(i, bits)
                distdata1 = Bgen.write(
                    file1,
                    distdata0,
                    bits=bits,
                    compression="zlib",
                    cleanup_temp_files=False,
                ).read()
                assert distdata1.iid[0, 0] == "0"
                distdata2 = Bgen(file1).read()
                assert distdata1.allclose(distdata2, equal_nan=True)
                atol = 1.0 / (2 ** (bits or 16))
                if (bits or 16) == 1:
                    atol *= 1.4  # because values must add up to 1, there is more rounding error
                TestBgen.assert_approx_equal(distdata0, distdata1, atol=atol)
        os.chdir(old_dir)

    def test_bad_sum(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        Path("temp").mkdir(parents=True, exist_ok=True)

        assert (
            "QCTOOLPATH" in os.environ
        ), "To run test_read_write_round_trip, QCTOOLPATH environment variable must be set. (On Windows, install QcTools in 'Ubuntu on Windows' and set to 'ubuntu run <qctoolLinuxPath>')."

        distdata = Bgen("../examples/example.bgen")[:5, :5].read()

        # Just one NaN
        distdata.val[0, 0, :] = [np.nan, 0.5, 0.5]
        bgen = Bgen.write("temp/should_be_all_nan.bgen", distdata)
        assert np.isnan(bgen[0, 0].read().val).all()

        # Just one NaN
        distdata.val[0, 0, :] = [0, 0, 0]
        bgen = Bgen.write("temp/should_be_all_nan2.bgen", distdata)
        assert np.isnan(bgen[0, 0].read().val).all()

        # Just sums to more than 1
        distdata.val[0, 0, :] = [1, 2, 3]
        failed = False
        try:
            bgen = Bgen.write("temp/should_fail.bgen", distdata)
        except Exception:
            failed = True
        assert failed

        # Just sums to less than 1
        distdata.val[0, 0, :] = [0.2, 0.2, 0.2]
        failed = False
        try:
            bgen = Bgen.write("temp/should_fail.bgen", distdata)
        except Exception:
            failed = True
        assert failed

        # a negative value
        distdata.val[0, 0, :] = [-1, 1, 0]
        failed = False
        try:
            bgen = Bgen.write("temp/should_fail.bgen", distdata)
        except Exception:
            failed = True
        assert failed
        os.chdir(old_dir)

    def test_read1(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        file_from = "../examples/example.bgen"
        file_to = "temp/example.bgen"
        pstutil.create_directory_if_necessary(file_to)
        if os.path.exists(file_to + ".metadata"):
            os.remove(file_to + ".metadata")
        meta = open_bgen._metadata_path_from_filename(file_to, samples_filepath=None)
        if os.path.exists(meta):
            os.remove(meta)
        shutil.copy(file_from, file_to)

        for loop_index in range(2):
            bgen = Bgen(file_to)
            assert np.array_equal(bgen.iid[0], ["0", "sample_001"])
            assert bgen.sid[0] == "SNPID_2,RSID_2"

            # Use the bgen_sample_id for both parts of iid
            def iid_dup(bgen_sample_id):
                return (bgen_sample_id, bgen_sample_id)

            iid_function = iid_dup
            bgen = Bgen(file_to, iid_function=iid_function, sid_function="id")
            assert np.array_equal(bgen.iid[0], ["sample_001", "sample_001"])
            assert bgen.sid[0] == "SNPID_2"

            bgen = Bgen(file_to, iid_function=iid_function, sid_function="rsid")
            assert np.array_equal(bgen.iid[0], ["sample_001", "sample_001"])
            assert bgen.sid[0] == "RSID_2"

            sid_function = lambda id, rsid: "{0},{1}".format(id, rsid)
            bgen = Bgen(file_to, iid_function, sid_function=sid_function)
            assert bgen.sid[0] == "SNPID_2,RSID_2"

        metafile = bgen._open_bgen._metadata_path_from_filename(
            file_to, samples_filepath=None
        )
        del bgen
        os.remove(metafile)
        sid_function = lambda id, rsid: "{0},{1}".format(id, rsid)
        bgen = Bgen(file_to, iid_function, sid_function=sid_function)
        assert bgen.sid[0] == "SNPID_2,RSID_2"

        metafile = bgen._open_bgen._metadata_path_from_filename(
            file_to, samples_filepath=None
        )
        del bgen
        os.remove(metafile)
        bgen = Bgen(file_to, iid_function, sid_function="rsid")
        assert np.array_equal(bgen.iid[0], ["sample_001", "sample_001"])
        assert bgen.sid[0] == "RSID_2"

        os.chdir(old_dir)

        # This and many of the tests based on bgen-reader-py\bgen_reader\test

    def test_bgen_samples_inside_bgen(self):
        with example_filepath("example.32bits.bgen") as filepath:
            data = Bgen(filepath)
            samples = [
                ("0", "sample_001"),
                ("0", "sample_002"),
                ("0", "sample_003"),
                ("0", "sample_004"),
            ]
            assert (data.iid[:4] == samples).all()

    def test_bgen_reader_variants_info(self):
        with example_filepath("example.32bits.bgen") as filepath:
            bgen = Bgen(filepath, sid_function="id")

            assert bgen.pos[0, 0] == 1
            assert bgen.sid[0] == "SNPID_2"
            assert bgen.pos[0, 2] == 2000

            assert bgen.pos[7, 0] == 1
            assert bgen.sid[7] == "SNPID_9"
            assert bgen.pos[7, 2] == 9000

            assert bgen.pos[-1, 0] == 1
            assert bgen.sid[-1] == "SNPID_200"
            assert bgen.pos[-1, 2] == 100001

            assert (bgen.iid[0] == ("0", "sample_001")).all()
            assert (bgen.iid[7] == ("0", "sample_008")).all()
            assert (bgen.iid[-1] == ("0", "sample_500")).all()

            g = bgen[0, 0].read()
            assert np.isnan(g.val).all()

            g = bgen[1, 0].read()
            a = [[[0.027802362811705648, 0.00863673794284387, 0.9635608992454505]]]
            np.testing.assert_array_almost_equal(g.val, a)

            b = [
                [
                    [
                        0.97970582847010945215516,
                        0.01947019668749305418287,
                        0.00082397484239749366197,
                    ]
                ]
            ]
            g = bgen[2, 1].read()
            np.testing.assert_array_almost_equal(g.val, b)

    def test_bgen_reader_without_metadata(self):
        with example_filepath("example.32bits.bgen") as filepath:
            bgen = Bgen(filepath)
            bgen.read()
            samples = bgen.iid
            assert samples[-1, 1] == "sample_500"

    def test_bgen_reader_file_notfound(self):
        bgen = Bgen("/1/2/3/example.32bits.bgen")
        try:
            bgen.iid  # expect error
            got_error = False
        except Exception:
            got_error = True
        assert got_error

    def test_bgen_reader_no_sample(self):
        with example_filepath("example.32bits.bgen") as filepath:
            bgen = Bgen(filepath)
            assert bgen.sid_count == 199

    def test_doctest(self):
        import doctest
        import pysnptools.distreader.bgen as bgen_mod

        old_level = logging.getLogger().level
        logging.getLogger().setLevel(logging.WARN)
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(bgen_mod)
        os.chdir(old_dir)
        logging.getLogger().setLevel(old_level)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_coverage(self):
        from pysnptools.distreader import DistGen

        with example_filepath("example.32bits.bgen") as filepath:
            bgen = Bgen(
                filepath, fresh_properties=False, iid_function=lambda sam: ("X", sam)
            )
            assert bgen.iid[0, 0] == "X"
            metadata_filepath = bgen._open_bgen._metadata2_path
            metadata2_temp = metadata_filepath.parent / (
                metadata_filepath.name + ".temp"
            )
            del bgen
            if metadata2_temp.exists():
                metadata2_temp.unlink()
            os.rename(metadata_filepath, metadata2_temp)
            bgen = Bgen(filepath)
            assert bgen.iid[0, 0] == "0"
            bgen[0, 0].read(order="A")
            if not os.path.exists("temp"):
                os.mkdir("temp")
            os.chdir("temp")
            file1x = "coverage.bgen"
            Bgen.write(file1x, bgen[:100, :100])
            Bgen.write(file1x, bgen[:100, :100])
            os.chdir("..")

        distgen0data = DistGen(seed=332, iid_count=10010, sid_count=5).read()
        file1 = "temp/roundtrip1-big.bgen"
        bed3 = Bgen.write(
            file1,
            distgen0data,
            bits=8,
            compression="zlib",
            cleanup_temp_files=False,
            sample_function=lambda fam, ind: f"{fam},{ind}",
        )
        bed3.iid[0, 0] = "0"


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestBgen))
    return test_suite


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.INFO)

    if False:
        os.chdir(r"D:\OneDrive\programs\pstsgkit\doc\ipynb")
        from pysnptools.distreader import DistData

        iid = [("0", "iid0"), ("0", "iid1")]
        sid = ["snp0", "snp1", "snp2"]
        pos = [
            [1, 0, 1],
            [1, 0, 2],
            [1, 0, 3],
        ]  # chromosome, genetic distance, basepair distance
        val = np.array(
            [
                [[0, 0, 1], [0.5, 0.25, 0.25], [0.95, 0.05, 0]],
                [[1, 0, 0], [44, 44, 44], [np.nan, np.nan, np.nan]],
            ],
            dtype="float32",
        )
        distdata = DistData(iid=iid, sid=sid, pos=pos, val=val, name="in-memory sample")
        distdata.val /= distdata.val.sum(axis=2, keepdims=True)
        # if you ask try to read a file that isn't there, do you get a sensible error?
        bgen = Bgen.write("2x3sample13.bgen", distdata, bits=23)  # write it
        bgen.read(dtype="float32").val  # Read the data from disk

    if False:
        import tracemalloc
        import logging
        import time

        logging.basicConfig(level=logging.INFO)
        tracemalloc.start()

        start = time.time()

        filename = "M:/deldir/genbgen/good/merged_487400x220000.bgen"
        # filename = "M:/deldir/genbgen/good/merged_487400x1100000.bgen"
        bgen = Bgen(filename, fresh_properties=False)
        val = bgen[:, 1000000:1000031].read().val
        # val = bgen[200000:200031, 100000:100031].read().val
        print("{0},{1:,}".format(val.shape, val.shape[0] * val.shape[1]))

        current, peak = tracemalloc.get_traced_memory()
        print(f"Current memory usage is {current / 10**6}MB; Peak was {peak / 10**6}MB")
        print("Time = {0} seconds".format(time.time() - start))
        tracemalloc.stop()

    if False:
        # filename = r"M:\deldir\fakeuk450000x1000.bgen"
        # filename = "M:/deldir/genbgen/good/merged_487400x220000.bgen"
        filename = "M:/deldir/genbgen/good/merged_487400x1100000.bgen"
        # filename = 'M:/deldir/genbgen/good/merged_487400x4840000.bgen'

        import tracemalloc
        import logging

        logging.basicConfig(level=logging.INFO)
        from pysnptools.distreader import Bgen

        tracemalloc.start()
        print(os.path.getsize(filename))
        bgen = Bgen(filename)
        print(bgen.shape)
        current, peak = tracemalloc.get_traced_memory()
        print(f"Current memory usage is {current / 10**6}MB; Peak was {peak / 10**6}MB")
        tracemalloc.stop()

    if False:
        logging.info("test info")
        from pysnptools.distreader import Bgen, DistGen

        for iid_count, sid_count in [(50, 5765294)]:
            print("iid_count=,sid_count=", iid_count, sid_count)
            dist_gen = DistGen(seed=332, iid_count=iid_count, sid_count=sid_count)
            filename = r"m:\deldir\fakeuk{0}x{1}.bgen".format(iid_count, sid_count)
            Bgen.write(
                filename, dist_gen, bits=8, compression="zlib", cleanup_temp_files=False
            )
            # print(os.path.getsize(filename))

    # if False: # !!!c,l
    #    from pysnptools.distreader import Bgen
    #    bgen = Bgen(r'D:\OneDrive\programs\hide\bgen-reader-py\bgen_reader\_example\complex.23bits.no.samples.bgen',allow_complex=True)
    #    print(bgen.sid_count)

    if False:
        from pysnptools.distreader import Bgen

        bgen = Bgen(r"M:\deldir\2500x100.bgen")
        bgen.read()
        print(bgen.shape)
        print("")

    if False:
        from pysnptools.distreader import Bgen

        bgen = Bgen(r"M:\deldir\1x1000000.bgen", verbose=True)
        print(bgen.shape)
        print("")

    if False:
        from pysnptools.distreader import Bgen

        bgen2 = Bgen(r"M:\deldir\10x5000000.bgen", verbose=True)
        print(bgen2.shape)

    if False:
        # iid_count = 500*1000
        # sid_count = 100
        # bits=8
        # #iid_count = 1
        # #sid_count = 1*1000*1000
        # iid_count = 2500
        # sid_count = 100
        # iid_count = 2500
        # sid_count = 500*1000
        # bits=16
        iid_count = 25
        sid_count = 1000
        bits = 16

        from pysnptools.distreader import DistGen
        from pysnptools.distreader import Bgen

        distgen = DistGen(seed=332, iid_count=iid_count, sid_count=sid_count)
        Bgen.write(
            r"M:\deldir\{0}x{1}.bgen".format(iid_count, sid_count), distgen, bits
        )
    if False:
        from pysnptools.distreader import Bgen

        bgen = Bgen(r"M:\deldir\500000x100.bgen")  # 1x1000000.bgen')
        print(bgen.iid)
        distdata = bgen.read(dtype="float32")
    if False:
        logging.basicConfig(level=logging.INFO)
        bgen = Bgen(
            r"M:\deldir\2500x500000.bgen", sid_function="id"
        )  # Bgen(r'M:\deldir\10x5000000.bgen')
        sid_index = int(0.5 * bgen.sid_count)
        distdata = bgen[:, sid_index].read()
        print(distdata.val)
    if False:
        from pysnptools.distreader import DistHdf5, Bgen

        distreader = DistHdf5("../examples/toydata.snpmajor.dist.hdf5")[
            :, :10
        ]  # A reader for the first 10 SNPs in Hdf5 format
        pstutil.create_directory_if_necessary("tempdir/toydata10.bgen")
        # Write data in BGEN format
        Bgen.write("tempdir/toydata10.bgen", distreader)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    import doctest

    logging.getLogger().setLevel(logging.WARN)
    result = doctest.testmod(optionflags=doctest.ELLIPSIS)
    logging.getLogger().setLevel(logging.INFO)
    assert result.failed == 0, "failed doc test: " + __file__



from setuptools import setup

# Version number
version = "0.5.11"


def readme():
    with open("README.md") as f:
        return f.read()


install_requires = [
    "scipy>=1.1.0",
    "numpy>=1.17.0",
    "pandas>=1.1.1",
    "psutil>=5.6.3",
    "h5py>=2.10.0",
    "cloudpickle>=2.2.0",
    "more-itertools>=8.7.0",
    "cbgen>= 1.0.4",
    "bgen-reader>=4.0.8",
    "bed-reader>=1.0.0",
]

# python setup.py sdist bdist_wininst upload
setup(
    name="pysnptools",
    version=version,
    description="PySnpTools",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords="gwas bioinformatics sets intervals ranges regions plink genomics "
    + "file-format reader genotype bed-format writer python snps",
    url="https://fastlmm.github.io/",
    author="FaST-LMM Team",
    author_email="fastlmm-dev@python.org",
    project_urls={
        "Bug Tracker": "https://github.com/fastlmm/PySnpTools/issues",
        "Documentation": "http://fastlmm.github.io/PySnpTools",
        "Source Code": "https://github.com/fastlmm/PySnpTools",
    },
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
    ],
    packages=[  # basically everything with a __init__.py
        "pysnptools",
        "pysnptools/kernelreader",
        "pysnptools/kernelstandardizer",
        "pysnptools/pstreader",
        "pysnptools/snpreader",
        "pysnptools/distreader",
        "pysnptools/standardizer",
        "pysnptools/util",
        "pysnptools/util/filecache",
        "pysnptools/util/mapreduce1",
        "pysnptools/util/mapreduce1/runner",
    ],
    package_data={
        "pysnptools": [
            "util/pysnptools.hashdown.json",
            "tests/mintest.py",
        ]
    },
    install_requires=install_requires,
)



import numpy as np
import scipy as sp
import logging
import doctest
import unittest
import os.path
import time
import sys
import tempfile
from pysnptools.pstreader import PstData, PstNpz, PstHdf5, PstMemMap
from pysnptools.util import create_directory_if_necessary
from pysnptools.kernelreader.test import _fortesting_JustCheckExists


class TestPstReader(unittest.TestCase):
    def test_big_npz(self):
        logging.info("in test_big_npz")
        n = 1000
        pstdata = PstData(
            row=range(n - 1), col=range(n + 1), val=np.zeros([n - 1, n + 1])
        )
        output = "tempdir/pstreader/big.npz"
        create_directory_if_necessary(output)
        PstNpz.write(output, pstdata)
        pstnpz = PstNpz(output)
        pstdata1 = pstnpz[::2, ::4].read()
        pstdata2 = pstnpz.read(order="A")
        assert pstdata2.val.flags["C_CONTIGUOUS"]

        pstdata = PstData(
            row=range(n - 1), col=range(n + 1), val=np.zeros([n - 1, n + 1], order="F")
        )
        PstNpz.write(output, pstdata)
        pstnpz = PstNpz(output)
        pstdata2 = pstnpz.read(order="A")
        pstdata2.val.flags["F_CONTIGUOUS"]

    def test_big_npz_3d(self):
        logging.info("in test_big_npz_3d")
        n = 1000
        pstdata = PstData(
            row=range(n - 1), col=range(n + 1), val=np.zeros([n - 1, n + 1, 3])
        )
        output = "tempdir/pstreader/big3d.npz"
        create_directory_if_necessary(output)
        PstNpz.write(output, pstdata)
        pstnpz = PstNpz(output)
        pstdata1 = pstnpz[::2, ::4].read()
        pstdata2 = pstnpz.read(order="A")
        assert pstdata2.val.flags["C_CONTIGUOUS"]

        pstdata = PstData(
            row=range(n - 1),
            col=range(n + 1),
            val=np.zeros([n - 1, n + 1, 3], order="F"),
        )
        PstNpz.write(output, pstdata)
        pstnpz = PstNpz(output)
        pstdata2 = pstnpz.read(order="A")
        pstdata2.val.flags["F_CONTIGUOUS"]

    def test_writes(self):
        # ===================================
        #    Defining sub functions
        # ===================================
        def _oned_int(c):
            return range(c)

        def _oned_str(c):
            return [str(i).encode("ascii") for i in range(c)]

        def _twooned_int(c):
            return [[i] for i in range(c)]

        def _twooned_str(c):
            return [[str(i).encode("ascii")] for i in range(c)]

        def _twotwod_int(c):
            return [[i, i] for i in range(c)]

        def _twotwod_str(c):
            return [[str(i).encode("ascii"), b"hello"] for i in range(c)]

        # def _twotwod_U(c):
        #    return [[str(i).encode('UTF-8'),u"hello"] for i in range(c)]
        def _none(c):
            return None

        def _zero(c):
            return np.empty([c, 0], dtype="S")

        # ===================================
        #    Starting main function
        # ===================================
        logging.info("starting 'test_writes'")
        np.random.seed(0)
        temp_dir = tempfile.TemporaryDirectory("pstreader")
        output_template = temp_dir.name + "/writes.{0}.{1}"
        i = 0
        for row_count in [5, 2, 1, 0]:
            for col_count in [4, 2, 1, 0]:
                for val_shape in [3, None, 1]:
                    val = (
                        np.random.normal(0.5, 2, size=(row_count, col_count))
                        if val_shape is None
                        else np.random.normal(
                            0.5, 2, size=(row_count, col_count, val_shape)
                        )
                    )
                    for row_or_col_gen in [
                        _oned_int,
                        _oned_str,
                        _twooned_int,
                        _twooned_str,
                        _twotwod_int,
                        _twotwod_str,
                    ]:  # !!!,_twotwod_U can't roundtrop Unicode in hdf5
                        row = row_or_col_gen(row_count)
                        col = row_or_col_gen(col_count)
                        for prop_gen in [
                            _none,
                            _oned_str,
                            _oned_int,
                            _twooned_int,
                            _twooned_str,
                            _twotwod_int,
                            _twotwod_str,
                            _zero,
                        ]:  # !!!_twotwod_U can't round trip Unicode because Hdf5 doesn't like it.
                            row_prop = prop_gen(row_count)
                            col_prop = prop_gen(col_count)
                            pstdata = PstData(row, col, val, row_prop, col_prop, str(i))
                            for the_class, suffix in [
                                (PstMemMap, "memmap"),
                                (PstHdf5, "hdf5"),
                                (PstNpz, "npz"),
                            ]:
                                filename = output_template.format(i, suffix)
                                logging.info(filename)
                                i += 1
                                the_class.write(filename, pstdata)
                                reader = (
                                    the_class(filename)
                                    if suffix != "hdf5"
                                    else the_class(filename, block_size=3)
                                )
                                _fortesting_JustCheckExists().input(reader)
                                for subsetter in [None, np.s_[::2, ::3]]:
                                    subreader = (
                                        reader
                                        if subsetter is None
                                        else reader[subsetter[0], subsetter[1]]
                                    )
                                    expected = (
                                        pstdata
                                        if subsetter is None
                                        else pstdata[subsetter[0], subsetter[1]].read()
                                    )
                                    for order in ["C", "F", "A"]:
                                        for force_python_only in [True, False]:
                                            readdata = subreader.read(
                                                order=order,
                                                force_python_only=force_python_only,
                                            )
                                            assert np.array_equal(
                                                readdata.val, expected.val
                                            )
                                            assert np.array_equal(
                                                readdata.row, expected.row
                                            )
                                            assert np.array_equal(
                                                readdata.col, expected.col
                                            )
                                            assert np.array_equal(
                                                readdata.row_property,
                                                expected.row_property,
                                            ) or (
                                                readdata.row_property.shape[1] == 0
                                                and expected.row_property.shape[1] == 0
                                            )
                                            assert np.array_equal(
                                                readdata.col_property,
                                                expected.col_property,
                                            ) or (
                                                readdata.col_property.shape[1] == 0
                                                and expected.col_property.shape[1] == 0
                                            )
                                if suffix in {"memmap", "hdf5"}:
                                    reader.flush()
                                os.remove(filename)
        temp_dir.cleanup()
        logging.info("done with 'test_writes'")

    def test_every_read(self):
        for order_from in ["F", "C"]:
            for order_to in ["F", "C"]:
                for dtype_from in [np.float32, np.float64]:
                    for dtype_to in [np.float32, np.float64]:
                        for val_shape in [None, 1, 3]:
                            for force_python_only in [True, False]:
                                np.random.seed(0)
                                val0 = (
                                    np.random.normal(0.5, 2, size=(3, 2))
                                    if val_shape is None
                                    else np.random.normal(
                                        0.5, 2, size=(3, 2, val_shape)
                                    )
                                )
                                val = np.array(val0, order=order_from, dtype=dtype_from)
                                pstdata = PstData(
                                    val=val, row=list(range(3)), col=list(range(2))
                                )
                                expected = np.array(
                                    val[::-2, :][:, ::-1],
                                    order=order_to,
                                    dtype=dtype_to,
                                )
                                result = pstdata[::-2, ::-1].read(
                                    order=order_to,
                                    dtype=dtype_to,
                                    force_python_only=force_python_only,
                                )
                                assert result.val.dtype == dtype_to
                                assert (
                                    order_to == "F" and result.val.flags["F_CONTIGUOUS"]
                                ) or (
                                    order_to == "C" and result.val.flags["C_CONTIGUOUS"]
                                )
                                assert np.array_equal(result.val, expected)

    def test_repr_test(self):
        np.random.seed(0)
        row_property = np.array([[1.0, 2, 2.5], [3, 4, 4.5], [5, 6, 6.5]])
        col_property = np.array([[1.0, 2, 2.5, 1], [3, 4, 4.5, 3]])
        pstdata = PstData(
            row=np.array([[1.0, 2], [3, 4], [5, 6]]),
            col=np.array([("A", "a"), ("B", "b")]),
            val=np.random.normal(0.5, 2, size=(3, 2)),
            row_property=row_property,
            col_property=col_property,
        )
        assert pstdata.col_to_index([("B", "b")])[0] == 1
        s = str(pstdata)

    def test_read(self):
        np.random.seed(0)
        row_property = np.array([[1.0, 2, 2.5], [3, 4, 4.5], [5, 6, 6.5]])
        col_property = np.array([[1.0, 2, 2.5, 1], [3, 4, 4.5, 3]])
        pstdata = PstData(
            row=np.array([[1.0, 2], [3, 4], [5, 6]]),
            col=np.array([["A", "a"], ["B", "b"]]),
            val=np.random.normal(0.5, 2, size=(3, 2)),
            row_property=row_property,
            col_property=col_property,
            name="test_read",
        )

        assert pstdata.row_to_index([np.array([3.0, 4])])[0] == 1
        assert pstdata.col_to_index([np.array(["A", "a"])])[0] == 0
        assert np.array_equal(pstdata[1:, :2].row_property, row_property[1:])
        assert np.array_equal(pstdata[1:, :2].col_property, col_property[:2])

        pstdata2 = pstdata[:2, :2].read()
        from pysnptools.kernelreader.test import _fortesting_JustCheckExists

        _fortesting_JustCheckExists().input(pstdata)
        _fortesting_JustCheckExists().input(pstdata2)

        np.testing.assert_array_almost_equal(
            pstdata2.val, pstdata.val[:2, :2], decimal=10
        )
        pstdata3 = pstdata[[], :].read()
        assert pstdata3.val.shape[0] == 0 and pstdata3.val.shape[1] == 2
        pstdata.val = pstdata.val.copy(order="F")
        pstdata2 = pstdata[:2, :2].read()
        np.testing.assert_array_almost_equal(
            pstdata2.val, pstdata.val[:2, :2], decimal=10
        )
        pstdata2 = pstdata[:2, :2].read(order="F")
        np.testing.assert_array_almost_equal(
            pstdata2.val, pstdata.val[:2, :2], decimal=10
        )
        pstdata2 = pstdata[:2, :2].read(order="A")
        np.testing.assert_array_almost_equal(
            pstdata2.val, pstdata.val[:2, :2], decimal=10
        )
        pstdata2 = pstdata[:2, :2].read(force_python_only=True, dtype=None, order="C")
        np.testing.assert_array_almost_equal(
            pstdata2.val, pstdata.val[:2, :2], decimal=10
        )
        pstdata2 = pstdata[:2, :2].read(
            force_python_only=True, dtype="float32", order="C"
        )
        np.testing.assert_array_almost_equal(
            pstdata2.val, pstdata.val[:2, :2].astype(dtype="float32"), decimal=10
        )
        pstdata2 = pstdata[:2, :2].read(
            force_python_only=True, dtype="float32", order=None
        )
        np.testing.assert_array_almost_equal(
            pstdata2.val, pstdata.val[:2, :2].astype(dtype="float32"), decimal=10
        )
        pstdata2 = pstdata[:2, :2].read(force_python_only=True, dtype=None, order="F")
        np.testing.assert_array_almost_equal(
            pstdata2.val, pstdata.val[:2, :2], decimal=10
        )
        pstdata4 = pstdata[::, ::].read(force_python_only=True)
        np.testing.assert_array_almost_equal(pstdata4.val, pstdata.val, decimal=10)

        logging.info("done with test")

    def test_inputs(self):
        from pysnptools.pstreader import PstData

        np.random.seed(0)
        row_property = np.array([1.0, 2, 2.5])
        col_property = np.array([1, 2])
        pstdata = PstData(
            row=np.array([1.0, 3, 6]),
            col=np.array(["Aa", "Bb"]),
            val=np.random.normal(0.5, 2, size=(3, 2)),
            row_property=row_property,
            col_property=col_property,
            name="test_read",
        )

        assert pstdata.row_to_index([3])[0] == 1
        assert pstdata.col_to_index(["Aa"])[0] == 0
        assert np.array_equal(pstdata[1:, :2].row_property, row_property[1:])
        assert np.array_equal(pstdata[1:, :2].col_property, col_property[:2])
        logging.info("done with test")

    def test_inputs2(self):
        from pysnptools.pstreader import PstData

        np.random.seed(0)
        row_property = None
        col_property = None
        pstdata = PstData(
            row=np.array([1.0, 3, 6]),
            col=np.array(["Aa", "Bb"]),
            val=np.random.normal(0.5, 2, size=(3, 2)),
            row_property=row_property,
            col_property=col_property,
            name="test_read",
        )

        assert pstdata.row_to_index([3])[0] == 1
        assert pstdata.col_to_index(["Aa"])[0] == 0
        assert np.array_equal(pstdata[1:, :2].row_property, pstdata.row_property[1:])
        assert np.array_equal(pstdata[1:, :2].col_property, pstdata.col_property[:2])
        logging.info("done with test")

    def test_inputs3(self):
        from pysnptools.pstreader import PstData

        np.random.seed(0)
        row_property = None
        col_property = None
        pstdata = PstData(
            row=[[1.0, 2.0], [3, 4], [6, 7]],
            col=np.array([]),
            val=[[], [], []],
            row_property=row_property,
            col_property=col_property,
            name="test_read",
        )

        assert pstdata.row_to_index([[3, 4]])[0] == 1
        assert np.array_equal(pstdata[1:, :2].row_property, pstdata.row_property[1:])
        assert np.array_equal(pstdata[1:, :2].col_property, pstdata.col_property[:2])
        logging.info("done with test")

    def test_flush(self):
        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        reader = PstHdf5("../examples/toydata.kernel.hdf5")
        val1 = reader[0, 0].read()
        reader.flush()
        val2 = reader[0, 0].read()
        assert val1 == val2

    def test_respect_read_inputs(self):
        from pysnptools.pstreader import _MergeRows, _MergeCols

        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        for pstreader in [
            PstNpz("../examples/toydata10.snp.npz"),
            _MergeRows(
                [
                    PstHdf5("../examples/toydata.snpmajor.snp.hdf5")[:5, :].read(),
                    PstHdf5("../examples/toydata.snpmajor.snp.hdf5")[5:, :].read(),
                ]
            ),
            _MergeCols(
                [
                    PstHdf5("../examples/toydata.snpmajor.snp.hdf5")[:, :5].read(),
                    PstHdf5("../examples/toydata.snpmajor.snp.hdf5")[:, 5:].read(),
                ]
            ),
            PstHdf5("../examples/toydata.snpmajor.snp.hdf5")[::2, ::2],
            PstHdf5("../examples/toydata.snpmajor.dist.hdf5").read(),
            PstHdf5("../examples/toydata.kernel.hdf5"),
            PstMemMap("../examples/tiny.pst.memmap"),
        ]:
            logging.info(str(pstreader))
            for order in ["F", "C", "A"]:
                for dtype in [np.float32, np.float64]:
                    for force_python_only in [True, False]:
                        for view_ok in [True, False]:
                            val = pstreader.read(
                                order=order,
                                dtype=dtype,
                                force_python_only=force_python_only,
                                view_ok=view_ok,
                            ).val
                            has_right_order = (
                                order == "A"
                                or (order == "C" and val.flags["C_CONTIGUOUS"])
                                or (order == "F" and val.flags["F_CONTIGUOUS"])
                            )
                            if hasattr(pstreader, "val") and not view_ok:
                                assert pstreader.val is not val
                            if (
                                hasattr(pstreader, "val")
                                and view_ok
                                and pstreader.val is not val
                                and (
                                    order == "A"
                                    or (
                                        order == "F"
                                        and pstreader.val.flags["F_CONTIGUOUS"]
                                    )
                                    or (
                                        order == "C"
                                        and pstreader.val.flags["C_CONTIGUOUS"]
                                    )
                                )
                                and (dtype is None or pstreader.val.dtype == dtype)
                            ):
                                logging.info(
                                    "{0} could have read a view, but didn't".format(
                                        pstreader
                                    )
                                )
                            assert val.dtype == dtype and has_right_order

        os.chdir(previous_wd)

    def test_inputs4(self):
        from pysnptools.pstreader import PstData

        pstdata = PstData(
            row=None,
            col=None,
            val=None,
            row_property=None,
            col_property=None,
            name="test_read",
        )

        assert (
            pstdata.row_count == 0
            and pstdata.col_count == 0
            and pstdata.val.shape[0] == 0
            and pstdata.val.shape[1] == 0
            and len(pstdata.row_property) == 0
            and len(pstdata.col_property) == 0
        )
        logging.info("done with test")


# We do it this way instead of using doctest.DocTestSuite because doctest.DocTestSuite requires modules to be pickled, which python doesn't allow.
# We need tests to be pickleable so that they can be run on a cluster.
class TestPstDocStrings(unittest.TestCase):
    pass

    def test_pstdata(self):
        import pysnptools.pstreader.pstdata

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.pstreader.pstdata, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_psthdf5(self):
        import pysnptools.pstreader.psthdf5

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.pstreader.psthdf5, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_pstmemmap(self):
        import pysnptools.pstreader.pstmemmap

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.pstreader.pstmemmap, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_pstnpz(self):
        import pysnptools.pstreader.pstnpz

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.pstreader.pstnpz, optionflags=doctest.ELLIPSIS
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_pstreader(self):
        import pysnptools.pstreader.pstreader

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(
            pysnptools.pstreader.pstreader,
            optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE,
        )
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_subset(self):
        np.random.seed(0)
        row_property = np.array([[1.0, 2, 2.5], [3, 4, 4.5], [5, 6, 6.5]])
        col_property = np.array([[1.0, 2, 2.5, 1], [3, 4, 4.5, 3]])
        val = np.random.normal(0.5, 2, size=(3, 2))
        pstdata = PstData(
            row=np.array([[1.0, 2], [3, 4], [5, 6]]),
            col=np.array([["A", "a"], ["B", "b"]]),
            val=val,
            row_property=row_property,
            col_property=col_property,
            name="test_read",
        )

        assert np.array_equal(pstdata[-1:0:-1, :].read().val, val[-1:0:-1, :])
        assert pstdata[-1, -1].read().val[0, 0] == val[-1, -1]
        assert np.array_equal(
            pstdata[[-1, 0], [-1, 0]].read().val, val[[-1, 0], :][:, [-1, 0]]
        )
        assert np.array_equal(
            pstdata[[True, False, True], [False, True]].read().val,
            val[[True, False, True], [False, True]].reshape(2, 1),
        )
        assert pstdata[0, 0].read().val[0, 0] == val[0, 0]
        assert np.array_equal(pstdata[1::2, 1::2].read().val, val[1::2, 1::2])

        logging.info("done with test")


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPstDocStrings))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPstReader))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()



import logging
import numpy as np
from pysnptools.kernelreader import KernelReader
from pysnptools.pstreader import PstHdf5
import warnings

class KernelHdf5(PstHdf5,KernelReader):
    '''
    A :class:`.KernelReader` for reading \*.kernel.hdf5 files from disk.

    See :class:`.KernelReader` for general examples of using KernelReaders.

    The general HDF5 format is described in `here <http://www.hdfgroup.org/HDF5/>`__. The KernelHdf5 format stores
    val, iid, sid, and pos information in Hdf5 format.
   
    **Constructor:**
        :Parameters: * **filename** (*string*) -- The KernelHdf5 file to read.

        :Example:

        >>> from pysnptools.kernelreader import KernelHdf5
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> hdf5_file = example_file('pysnptools/examples/toydata.kernel.hdf5')
        >>> data_on_disk = KernelHdf5(hdf5_file)
        >>> print(data_on_disk.iid_count)
        500

    **Methods beyond** :class:`.KernelReader`

    '''
    def __init__(self,*args, **kwargs):
        super(KernelHdf5, self).__init__(*args, **kwargs)

    @property
    def row(self):
        self._run_once()
        if self._row.dtype.type is not np.str_:
            if self._row is self._col:
                self._row = np.array(self._row,dtype='str')
                self._col = self._row
            else:
                self._row = np.array(self._row,dtype='str')
        return self._row

    @property
    def col(self):
        self._run_once()
        if self._col.dtype.type is not np.str_:
            if self._row is self._col:
                self._col = np.array(self._col,dtype='str')
                self._row = self._col
            else:
                self._col = np.array(self._col,dtype='str')
        return self._col


    @staticmethod
    def write(filename, kerneldata, hdf5_dtype=None, sid_major=True):
        """Writes a :class:`KernelData` to KernelHdf5 format and returns the :class:`.KernelHdf5`.

        :param filename: the name of the file to create
        :type filename: string
        :param kerneldata: The in-memory data that should be written to disk.
        :type kerneldata: :class:`KernelData`
        :param hdf5_dtype: None (use the .val's dtype) or a Hdf5 dtype, e.g. 'f8','f4',etc.
        :type hdf5_dtype: string
        :param col_major: Tells if vals should be stored on disk in sid_major (default) or iid_major format.
        :type col_major: bool
        :rtype: :class:`.KernelHdf5`

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.standardizer import Unit
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file('pysnptools/examples/toydata.5chrom.*','*.bed')
        >>> kerneldata = Bed(bed_file,count_A1=False).read_kernel(Unit())     # Create a kernel from the data in the Bed file
        >>> pstutil.create_directory_if_necessary("tempdir/toydata.kernel.hdf5")
        >>> KernelHdf5.write("tempdir/toydata.kernel.hdf5",kerneldata)          # Write data in KernelHdf5 format
        KernelHdf5('tempdir/toydata.kernel.hdf5')
        """
        PstHdf5.write(filename,kerneldata,hdf5_dtype=hdf5_dtype,col_major=sid_major)
        return KernelHdf5(filename)


if __name__ == "__main__":

    logging.basicConfig(level=logging.INFO)

    if False:
        from pysnptools.kernelreader import KernelHdf5
        data_on_disk = KernelHdf5('../examples/toydata.kernel.hdf5')
        print(data_on_disk.iid_count)


    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
import subprocess, sys, os.path
from itertools import *
import pandas as pd
import logging
from pysnptools.pstreader import PstReader
from pysnptools.pstreader.pstdata import PstData
import pysnptools.util as pstutil
import warnings

class _OneShot(PstReader):
    '''
    A utility class that makes it easier to define readers that read all the data and into memory before doing any subsetting.
    '''

    def _read_pstdata():
        raise NotImplementedError("{0} needs to define its own _read_pstdata".format(self.__class__.__name__))

    def __init__(self):
        super(_OneShot, self).__init__()

        self._ran_once = False

    def __repr__(self): 
        if hasattr(self,"filename"):
            return "{0}('{1}')".format(self.__class__.__name__,self.filename)
        else:
            return "{0}".format(self.__class__.__name__)

    def copyinputs(self, copier):
        if hasattr(self,"filename"):
            # doesn't need to self._run_once()
            copier.input(self.filename)

    def _run_once(self):
        if (self._ran_once):
            return
        self._ran_once = True
        self._data = self._read_pstdata()

    @property
    def row(self):
        self._run_once()
        return self._data._row

    @property
    def col(self):
        self._run_once()
        return self._data._col

    @property
    def row_property(self):
        self._run_once()
        return self._data._row_property

    @property
    def col_property(self):
        self._run_once()
        return self._data._col_property

    # Most _read's support only indexlists or None, but this one supports Slices, too.
    _read_accepts_slices = True
    def _read(self, row_index_or_none, col_index_or_none, order, dtype, force_python_only, view_ok, num_threads):
        self._run_once()
        dtype = np.dtype(dtype)
        val = self._data._read(row_index_or_none, col_index_or_none, order, dtype, force_python_only, view_ok, num_threads)
        return val




from __future__ import absolute_import
import os
import cloudpickle
import pysnptools.util as pstutil
from itertools import *
from unittest.mock import patch


def work_sequence_to_result_sequence(work_sequence):
    """
    Does all the work in a sequence of work items and returns a sequence of results.
    """
    for work in work_sequence:
        if callable(work):
            result = work()
        else:
            result = _run_all_in_memory(work)
        yield result


def _run_all_in_memory(work):
    """
    If callable, just calls
    If distributable, does all the work and then calls reduce on the results.
    """
    if callable(work):
        return work()
    else:
        assert hasattr(
            work, "work_sequence"
        ), "Your work item doesn't have a work_sequence. This can be caused by having 'map_reduce(nested=something,...)' where you should have 'map_reduce(mapper=something,...)' because the 'something' is not a nested 'map_reduce'. The exception can also be caused by giving a 'runner' on an inner, nested map_reduce."
        work_sequence = work.work_sequence()
        result_sequence = work_sequence_to_result_sequence(work_sequence)
        return work.reduce(result_sequence)


def create_task_file_name(workdirectory, taskindex, taskcount):
    """
    Creates a file name to use for saving a set of results.
    """
    task_file_name = os.path.join(
        workdirectory, "{0}.{1}.p".format(taskindex, taskcount)
    )
    return task_file_name


def _shape_to_desired_workcount(distributable, taskcount, weights=None):
    workcount = distributable.work_count
    if workcount > taskcount:
        distributable = BatchUpWork(distributable, workcount, taskcount, weights)
    else:
        if workcount < taskcount:
            if workcount == 0:
                distributable = MakeWork(distributable, taskcount)
            else:
                distributable = ExpandWork(distributable, workcount, taskcount)
    return distributable


def _run_one_task(
    original_distributable,
    taskindex,
    taskcount,
    workdirectory,
    weights=None,
    environ=None,
):
    """
    Does a fraction of the work (e.g. 1 of every 1000 work items) and then saves the results to single file.
    if taskindex == taskcount, does the reduce step
    """

    if not 0 < taskcount:
        raise Exception("Expect taskcount to be positive")
    if not (0 <= taskindex < taskcount + 1):
        raise Exception(
            "Expect taskindex to be between 0 (inclusive) and taskcount (exclusive)"
        )

    shaped_distributable = _shape_to_desired_workcount(
        original_distributable, taskcount, weights=weights
    )

    if shaped_distributable.work_count != taskcount:
        raise Exception("Assert: expect workcount == taskcount")

    pstutil.create_directory_if_necessary(workdirectory, isfile=False, robust=True)

    with patch.dict("os.environ", environ if environ is not None else {}) as _:
        if taskindex < taskcount:
            doMainWorkForOneIndex(
                shaped_distributable, taskcount, taskindex, workdirectory
            )
            return None
        else:
            result_sequence = work_sequence_from_disk(workdirectory, taskcount)
            return shaped_distributable.reduce(result_sequence)


def _work_sequence_for_one_index(distributable, taskAndWorkcount, taskindex):
    if hasattr(distributable, "work_sequence_range"):
        is_first_and_only = True
        for work in distributable.work_sequence_range(taskindex, taskindex + 1):
            assert is_first_and_only, "real assert"
            is_first_and_only = False
            yield work
    else:
        for workIndex, work in enumerate(distributable.work_sequence()):
            if workIndex == taskAndWorkcount:
                raise Exception(
                    "Expect len(work_sequence) to match work_count, but work_sequence was too long"
                )
            if workIndex == taskindex:
                yield work
                workDone = True
                if (
                    workIndex != taskAndWorkcount - 1
                ):  # the work is done, so quit enumerating work (but don't quit early if you're the last workIndex because we want to double check that the work_sequence and work_count match up)
                    break
        if not workDone:
            raise Exception(
                "Expect len(work_sequence) to match work_count, but work_sequence was too short"
            )


def doMainWorkForOneIndex(distributable, taskAndWorkcount, taskindex, workdirectory):
    task_file_name = create_task_file_name(workdirectory, taskindex, taskAndWorkcount)
    workDone = False
    with open(task_file_name, mode="wb") as f:
        for work in _work_sequence_for_one_index(
            distributable, taskAndWorkcount, taskindex
        ):
            result = _run_all_in_memory(work)
            cloudpickle.dump(result, f)  # save a result to the temp results file


def work_sequence_from_disk(workdirectory, taskAndWorkcount):
    """
    Reads all the result sets from the temporary files and returns a sequence results
    """
    for taskindex in range(taskAndWorkcount):
        task_file_name = create_task_file_name(
            workdirectory, taskindex, taskAndWorkcount
        )
        with open(task_file_name, mode="rb") as f:
            try:
                result = cloudpickle.load(f)
            except Exception as detail:
                raise Exception(
                    "Error trying to unpickle '{0}'. {1}".format(task_file_name, detail)
                )
        # if True:
        #    logging.debug("\tAbout to yield result {0} of {1}".format(taskindex,taskAndWorkcount))
        yield result


class BatchUpWork(object):  # implements IDistributable
    """
    A wrapper.
    Takes a distributable that has more work items that wanted and turns it into one with exactly the right number of work items.
    It does this by batching up work items, block style.

    If weights is given, it must be an array of integers of length "taskcount". Each value gives the relative weight
    of that task compared to the others. For example, if all values are 1, then all task will be assigned
    the same amount of work. If the first value is 2 and the others are 1, then the first task (index 0)
    will be assigned twice as much work as the other tasks.
    """

    def __init__(self, distributable, workcount, taskcount, weights=None):
        self.sub_distributable = distributable
        self.sub_workcount = workcount
        self._workcount = taskcount

        if taskcount <= 0:
            raise Exception("Expect taskcount to be at least 1.")

        if weights is None:
            self._partial_sum_weight = list(range(0, taskcount + 1))
        else:
            if len(weights) != taskcount:
                raise Exception("Expect weights to have length of taskcount.")
            self._partial_sum_weight = [0]
            for weight in weights:
                if weight != int(weight) or weight < 0:
                    raise Exception("Expect weights be non-negative integers.")
                self._partial_sum_weight.append(weight + self._partial_sum_weight[-1])
        if self._partial_sum_weight[-1] == 0:
            raise Exception("Weights cannot be all zero.")

    @property
    def work_count(self):
        return self._workcount

    def work_sequence(self):
        return work_sequence_range(0, self._workcount)

    def work_sequence_range(self, start, stop):
        assert 0 <= start <= stop <= self._workcount, "real assert"
        for workIndex in range(start, stop):
            yield lambda workIndex=workIndex: self.work(workIndex)

    def work(self, workIndex):
        result = [workIndex]
        start, stop = self.createSubWorkIndexList(workIndex)
        for sub_workIndex, sub_work in self.pull_sub_work(start, stop):
            sub_result = _run_all_in_memory(sub_work)
            result.append((sub_workIndex, sub_result))
        return result

    def pull_sub_work(self, start, stop):
        if hasattr(self.sub_distributable, "work_sequence_range"):
            index = start
            for sub_work in self.sub_distributable.work_sequence_range(start, stop):
                yield index, sub_work
                index += 1
            assert index == stop, "real assert"
        else:
            sub_workIndexList = list(range(stop - 1, start - 1, -1))
            for sub_workIndex, sub_work in enumerate(
                self.sub_distributable.work_sequence()
            ):
                if sub_workIndex == self.sub_workcount:
                    raise Exception("Expect len(work_sequence) to match work_count")
                if sub_workIndex == sub_workIndexList[-1]:
                    yield sub_workIndex, sub_work
                    sub_workIndexList.pop()
                    if (
                        len(sub_workIndexList) == 0
                        and sub_workIndex != self.sub_workcount - 1
                    ):  # there are no more work items for this task, so quit (but don't quit early if you're the last workIndex because we want to double check that the work_sequence and work_count match up)
                        break
            if len(sub_workIndexList) != 0:
                raise Exception(
                    "no work was enumerated for workIndex={0}".format(
                        sub_workIndexList[-1]
                    )
                )

    def reduce(self, result_sequence):
        sub_result_sequence = self.create_sub_result_sequence(result_sequence)
        result = self.sub_distributable.reduce(sub_result_sequence)
        return result

    def create_sub_result_sequence(self, result_sequence):
        for result in result_sequence:
            workIndex = result.pop(0)
            start, stop = self.createSubWorkIndexList(workIndex)
            if (stop - start) != len(result):
                raise Exception("Assert: batched results not expected size")
            for sub_workIndex, pair in zip(range(start, stop), result):
                sub_workIndex_check, sub_result = pair
                if sub_workIndex != sub_workIndex_check:
                    raise Exception("Assert: Unexpected workindex in batched result")
                yield sub_result

    @property
    def tempdirectory(self):
        return self.sub_distributable.tempdirectory

    # optional
    def __str__(self):
        return "{0}({1},{2},{3})".format(
            self.__class__.__name__,
            self.sub_distributable,
            self.sub_workcount,
            self._workcount,
        )

    def createSubWorkIndexList(self, workindex):
        assert 0 <= workindex < self._workcount, "real assert"
        start = (
            self._partial_sum_weight[workindex]
            * self.sub_workcount
            // self._partial_sum_weight[-1]
        )  # assuming high precision integer math
        stop = (
            self._partial_sum_weight[workindex + 1]
            * self.sub_workcount
            // self._partial_sum_weight[-1]
        )  # assuming high precision integer math
        return start, stop


class ExpandWork(object):  # implements IDistributable
    def __init__(self, distributable, workcount, taskcount):
        assert 1 <= workcount, "Expect workcount to be at least 1"
        assert (
            workcount < taskcount
        ), "ExpandWork expects workcount to be less than taskcount"
        self.sub_distributable = distributable
        self.sub_workcount = workcount
        self._workcount = taskcount
        self.floor = taskcount // workcount  # all work gets this many tasks
        self.excess = (
            taskcount - self.floor * workcount
        )  # but there are this many extra tasks
        self.extra_index = (
            self.floor + 1
        ) * self.excess  # tasks before here, get one extra task

        if hasattr(distributable, "work_sequence_range"):
            self.work_sequence_range = self._work_sequence_range

    @property
    def work_count(self):
        return self._workcount

    def expand_to(self, sub_workIndex):
        if sub_workIndex < self.excess:
            return self.floor + 1
        else:
            return self.floor

    def index_to_sub_index(self, index):
        if index > self.extra_index:
            sub_index = self.excess + (index - self.extra_index) // self.floor
        else:
            sub_index = index // (self.floor + 1)
        return sub_index

    def sub_index_to_start(self, sub_index):
        if sub_index < self.excess:
            start = sub_index * (self.floor + 1)
        else:
            start = self.extra_index + (sub_index - self.excess) * self.floor
        return start

    def sub_sub_start_stop(self, sub_start, start, stop):
        full_start = self.sub_index_to_start(sub_start)
        full_end = self.sub_index_to_start(sub_start + 1)
        if full_start < start:
            sub_sub_start = start - full_start
        else:
            sub_sub_start = 0

        if stop < full_end:
            sub_sub_stop = stop - full_start
        else:
            sub_sub_stop = full_end - full_start

        return sub_sub_start, sub_sub_stop

    def sub_distributable_work_sequence_range(self, start, stop):
        if hasattr(self.sub_distributable, "work_sequence_range"):
            return self.sub_distributable.work_sequence_range(start, stop)
        else:
            return islice(self.sub_distributable.work_sequence(), start, stop)

    def _work_sequence_range(self, start, stop):
        sub_start = self.index_to_sub_index(start)
        sub_stop = self.index_to_sub_index(stop - 1) + 1
        sub_workIndex = sub_start
        workIndex = start
        for sub_work in self.sub_distributable_work_sequence_range(sub_start, sub_stop):
            expand_to = self.expand_to(sub_workIndex)
            sub_sub_start, sub_sub_stop = self.sub_sub_start_stop(
                sub_workIndex, start, stop
            )
            if not callable(
                sub_work
            ):  # i.e. a distributable job. If our expand_to is 3 and i's work_count is 9, then it is batched up 3x3. On the other hand, if its work_count is 2, then we ExpandWork from 2 to 3.
                shaped_distributable = _shape_to_desired_workcount(sub_work, expand_to)
                for sub_sub_work in shaped_distributable.work_sequence_range(
                    sub_sub_start, sub_sub_stop
                ):
                    yield sub_sub_work
                    workIndex += 1
                #    expand_to -= 1
                # if expand_to != 0 :
                #    raise Exception("Assert: shaped_distributable doesn't have right number of work items")
                # assert sub_sub_index == sub_sub_end, "real assert"
            else:
                sub_sub_index = sub_sub_start
                if sub_sub_index == 0:
                    yield sub_work
                    workIndex += 1
                    sub_sub_index += 1
                for sub_sub_index2 in range(sub_sub_index, sub_sub_stop):
                    assert sub_sub_index2 == sub_sub_index, "real assert"
                    yield lambda: None
                    workIndex += 1
                    sub_sub_index += 1
                assert sub_sub_index == sub_sub_stop, "real assert"
            sub_workIndex += 1
        assert sub_workIndex == sub_stop, "real assert"
        assert workIndex == stop, "real assert"

    def work_sequence(self):
        return self._work_sequence_range(0, self.work_count)

    def reduce(self, result_sequence):
        sub_result_sequence = self.create_sub_result_sequence(result_sequence)
        result = self.sub_distributable.reduce(sub_result_sequence)
        return result

    def create_sub_result_sequence(self, result_sequence):
        # e.g. we had 11 work items, but we wanted 300 so each work item was expanded to 28 or 27 work items.
        # now we have 300 results that we must turn to 11 results.
        # If we had just padded with empty work we could just filter those out and we'd have 11. But we might
        # have gone down another level of map reduce.
        # We assume that the 300 items must arrive in order.
        sub_workIndex = 0
        workIndex = 0
        for sub_work in self.sub_distributable.work_sequence():
            expand_to = self.expand_to(sub_workIndex)
            if not callable(sub_work):
                shaped_distributable = _shape_to_desired_workcount(sub_work, expand_to)
                sub_result = SubGen(result_sequence, expand_to)
                result = shaped_distributable.reduce(sub_result)
                if not hasattr(shaped_distributable, "suppress_yield"):
                    yield result
                workIndex += expand_to
            else:
                yield next(result_sequence)
                workIndex += 1
                for workindex in range(1, expand_to):
                    paddedResultToIgnore = next(result_sequence)
                    if paddedResultToIgnore is not None:
                        raise Exception("Assert: Expected 'None' result")
                    workIndex += 1
            sub_workIndex += 1
        if sub_workIndex != self.sub_workcount:
            raise Excpetion("Assert:  expect len(result_sequence) to match workcount")
        if workIndex != self._workcount:
            raise Exception("Assert: expect len(result_sequence) to match workcount")
        try:
            next(
                result_sequence
            )  # should get an StopIternation here, which will be ignored.
            raise Exception(
                "Assert: expect len(result_sequence) to match workcount. This can be caused by a 'reducer' that doesn't pull every input from its input sequence."
            )
        except StopIteration:
            pass  # do nothing

    @property
    def tempdirectory(self):
        return self.sub_distributable.tempdirectory

    # optional
    def __str__(self):
        return "{0}({1},{2},{3})".format(
            self.__class__.__name__,
            self.sub_distributable,
            self.sub_workcount,
            self._workcount,
        )


class MakeWork(object):  # implements IDistributable
    def __init__(self, distributable, taskcount):
        self._workcount = taskcount
        self.sub_distributable = distributable

    suppress_yield = None  # Just defining it is enough. The value doesn't matter.

    @property
    def work_count(self):
        return self._workcount

    def work_sequence_range(self, start, stop):
        for i in range(start, stop):
            yield lambda: None

    def work_sequence(self):
        return self._work_sequence_range(0, self.work_count)

    def reduce(self, result_sequence):
        for (
            result
        ) in result_sequence:  # We iterate through, just so everything is accounted for
            pass
        return "This return value should always be ignored"

    def __str__(self):
        return "{0}({1})".format(self.__class__.__name__, self._workcount)


class SubGen:
    def __init__(self, gen, count):
        self.gen = gen
        self.count = count

    def __iter__(self):
        return self

    def __next__(self):
        if self.count > 0:
            self.count -= 1
            return next(self.gen)
        else:
            raise StopIteration()


from pysnptools.util.mapreduce1.runner.runner import Runner
from pysnptools.util.mapreduce1.runner.local import Local, _JustCheckExists
from pysnptools.util.mapreduce1.runner.localmultiproc import LocalMultiProc
from pysnptools.util.mapreduce1.runner.localmultithread import LocalMultiThread
from pysnptools.util.mapreduce1.runner.localinparts import LocalInParts



import os
import logging
import unittest
from pysnptools.util.filecache import Hashdown
import doctest

pysnptools_hashdown = Hashdown.load_hashdown(
    os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "pysnptools.hashdown.json"
    ),
    directory=os.environ.get("PYSNPTOOLS_CACHE_HOME", None),
)

def example_file(pattern, endswith=None):
    """
    Returns the local location of a PySnpTools example file, downloading it
    if needed.

    :param pattern: The name of the example file of interest. A 
        `file name pattern <https://docs.python.org/3.7/library/fnmatch.html>`__
        may be given. All matching files will be downloaded (if needed) and
        the name of one will be returned.
    :type file_name: string

    :param endswith: The pattern of the file name to return. By default, when
        no `endswith` is given, the name of the first matching file will be
        returned.
    :type file_name: string

    :rtype: string

    By default, the local location will be under the system temp directory
    (typically controlled with the TEMP environment variable).
    Alternatively, the directory can be set with the PYSNPTOOLS_CACHE_HOME
    environment variable.

    This function knows the MD5 hash of all PySnpTools example files and uses
    that content-based hash to decide if a file needs to be downloaded.

    >>> from pysnptools.util import example_file # Download and return local file name
    >>> # Download the phenotype file if necessary. Return its local location.
    >>> pheno_fn = example_file("pysnptools/examples/toydata.phe")
    >>> print('The local file name is ', pheno_fn)
    The local file name is ...pysnptools/examples/toydata.phe
    >>>
    >>> # Download the bed,bim,&fam files if necessary. Return the location of bed file.
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> print('The local file name is ', bedfile)
    The local file name is ...tests/datasets/all_chr.maf0.001.N300.bed

    """
    return pysnptools_hashdown._example_file(pattern, endswith=endswith)


class TestExampleFile(unittest.TestCase):
    def test_doc_test(self):
        import pysnptools.util._example_file as example_mod

        result = doctest.testmod(
            example_mod, optionflags=doctest.ELLIPSIS
        )
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    """
    set up test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestExampleFile))

    return test_suite


if __name__ == "__main__":
    # This creates a Json based on all files, but it may have the wrong line-endings and thus hash on Windows. It can be edited to files of interest
    if True:
        update = Hashdown(url=pysnptools_hashdown.url, file_to_hash=pysnptools_hashdown.file_to_hash, allow_unknown_files=True)
        for file0 in list(os.walk(r'D:\OneDrive\programs\pysnptools\pysnptools\examples'))[0][2:][0]:
            file = 'pysnptools/examples/'+file0
            print(file)
            update.file_exists(file)
        update.save_hashdown("deldir/updated.hashdown.json")
    if False:
        Hashdown.scan_local(
            r"D:\OneDrive\programs\pysnptools",
            url="https://github.com/fastlmm/PySnpTools/raw/cf248cbf762516540470d693532590a77c76fba2",
        ).save_hashdown("deldir/pysnptools.hashdown.json")
    # This fixes the hashes of the files
    if False:
        update = Hashdown(url=pysnptools_hashdown.url, allow_unknown_files=True)
        for file in pysnptools_hashdown.file_to_hash:
            print(file)
            update.file_exists(file)
        update.save_hashdown("deldir/updated.hashdown.json")

    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    doctest.testmod(optionflags=doctest.ELLIPSIS)
